#!/usr/bin/env python3.9
# -*- coding: utf-8 -*- 

# Copyright (c) 2021 Marco Ferrara

# License:
# GNU GPL License
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Dependencies: 
# TODO...

import sys, os
import site
import re
import bpy
import inspect
import time
import shutil
import subprocess
import itertools
from pathlib import Path as Filepath
cwd = os.path.dirname(os.path.realpath(__file__))
site.addsitedir(os.path.join(cwd, "libs", "site", "packages"))

from prj.drawing_camera import get_drawing_camera
from prj.drawing_maker import draw, tmp_svgs
from prj.drawing_subject import get_subjects_list, reset_subjects_list
from prj.selection_subjects import get_selected_subjects
from prj.subjects_getter import get_in_frame_subjects
from prj.subjects_processor import get_overlaps
from prj.svg_data import reset_svgs_data
from prj.svg_utils import subjects_in_svg, get_path_coords, get_abs_svg
from prj.svg_utils import transform_points, join_coords
from prj.svglib import AbsGroup, AbsLayer, AbsPath
from prj.svgread import Svg_read
from prj.utils import get_resolution, flatten, is_renderable, set_hi_res_percent
from prj.utils import add_modifier, debug_here, get_render_basepath, tmp_imgs
from prj.utils import remove_file, check_dir, get_addon_preferences
from prj.utils import get_drawing_styles, UNITS_SYSTEM
from prj.geometry_utils import is_cut
from prj.visibility_checker import get_not_occluded_subjects
from prj.working_scene import get_working_scene
from prj.cad import generate_dxf
from prj.progress_indicator import get_progress_indicator


BASE_CSS = 'prj.css'
PRJ_DIR = Filepath(__file__).parent.absolute()
ASSETS_DIR = Filepath(PRJ_DIR / 'assets')
RESOLUTION_FACTOR: float = 96.0 / 2.54 ## resolution / inch
ROUNDING: int = 5
drawing_styles = None
get_css_style = lambda location: f"@import url({location}{BASE_CSS});"
to_be_drawn = lambda subj: subj.xray_drawing or subj.force_drawing \
        or subj.is_symbol or subj.back_drawing
sorted_dict = lambda ds:sorted(ds, key=lambda x: ds[x].position)
main_progress_count = itertools.count()
re_stroke_width = re.compile(r'(stroke-width\s*:)\s*([\.\d]*)')
re_stroke_dasharray = re.compile(r'(stroke-dasharray\s*:)([\s\.\d]*)')


def set_subjects_visibility(subject: 'Drawing_subject') -> None:
    """ Hide and show subjects depending on situation """
    subject.obj.hide_viewport = False
    if subject.has_back_stairs():
        subject.back_stairs_obj.hide_viewport = False
    drawing_subjects = get_subjects_list()
    for other_subj in drawing_subjects:
        if other_subj == subject:
            continue
        if other_subj.has_back_stairs():
            other_subj.back_stairs_obj.hide_viewport = True
            continue
        if other_subj.xray_drawing:
            other_subj.obj.hide_viewport = True
            continue
        if other_subj not in subject.overlapping_subjects:
            other_subj.obj.hide_viewport = True
            continue
        other_subj.obj.hide_viewport = False

def prepare_subjects(subjects: list['Drawing_subject']) -> None:
    """ Prepare cut objects, simplify meshes and apply stairs cuts """

    print('Prepare drawings')
    prepare_start_time = time.time()
    working_scene = get_working_scene()
    drawing_camera = get_drawing_camera()

    ## Hide frame box (which cuts out cut objects to frame boundary)
    drawing_camera.frame_box.hide_set(True)

    to_cut = []
    to_simplify = []
    to_stairs_cut = []
    for i, subject in enumerate(subjects):
        if subject.is_symbol:
            if not subject.is_stairs_cut:
                continue
            to_stairs_cut.append(subject)
        if subject.is_cut:
            to_cut.append(subject)
        if subject.hidden_drawing:
            continue
        to_simplify.append(subject)

    progress = get_progress_indicator()
    progress.set_phase(phase_data={
        'name': 'prepare_subjects', 'start':progress.get_last_update(), 
        'end':progress.get_next_step('main'), 
        'steps':len(to_cut + to_simplify + to_stairs_cut)})

    j = 0
    for i, subject in enumerate(to_stairs_cut): 
        progress.update(phase_name='prepare_subjects', step=i)
        print('Apply stairs cut for', subject.name)
        subject.apply_stairs_cut()
        j += 1

    for i, subject in enumerate(to_cut, start=j):
        progress.update(phase_name='prepare_subjects', step=i)
        if subject.stairs_cut_ref:
            continue
        print('Create cut for', subject.name)
        cut_obj = subject.create_cut()
        add_modifier(cut_obj, 'prj_frame_cut_bool', 'BOOLEAN',
                {'object': drawing_camera.frame_box, 
                    'operation': 'INTERSECT', 'solver': 'FAST'})
        if cut_obj.name not in working_scene.scene.collection.objects:
            working_scene.link_object(cut_obj)
        j += 1
        #if subject.name == 'wall_037':
        #    debug_here()

    ## TODO Remove faces behind camera too (if not back drawing)
    for i, subject in enumerate(to_simplify, start=j):
        progress.update(phase_name='prepare_subjects', step=i)
        print('Simplify mesh for', subject.name)
        subject.get_visible_mesh()

    preparing_time = time.time() - prepare_start_time
    print(f"Preparing subjects in {preparing_time} seconds")

def draw_subjects(subjects: list['Drawing_subject']) -> None:
    """ Get exported svgs for every subject (or parts of it) for every style """
    drawing_times: dict[float, str] = {}
    draw_time = time.time()
    ## Resolution percentage affects grease pencil svg export: reset it to 100
    get_working_scene().set_resolution_percentage(100)

    progress = get_progress_indicator()
    progress.set_phase(phase_data={'name': 'draw_subjects',
                'start':progress.get_last_update(), 
                'end':progress.get_next_step('main'), 'steps':len(subjects) })

    ## Draw every subject (and hide not overlapping ones)
    for i, subject in enumerate(subjects):
        print('Drawing', subject.name)
        progress.update(phase_name='draw_subjects', step=i)

        drawing_start_time = time.time()
        set_subjects_visibility(subject)
        draw(subject)

        ## It misses same-time drawing objects
        drawing_time = time.time() - drawing_start_time
        drawing_times[drawing_time] = subject.name
        print(f"\t...drawn in {drawing_time} seconds")

    draw_time = time.time() - draw_time
    print('\n')
    for t in sorted(drawing_times):
        print(drawing_times[t], t)
    print(f"***Drawings completed in {draw_time} seconds")
    return 

def collect_subject_svgs(subjects: list['Drawing_subject']) -> None:
    """ Get a single and organized svg for every subject and join cut, prj, 
        bak... """
    print('Start rewriting svg')
    rewrite_svgs_start_time = time.time()

    from prj.svg_data import svgs_data

    progress = get_progress_indicator()
    progress.set_phase(phase_data={'name': 'collect_subject_svgs',
                'start':progress.get_last_update(), 
                'end':progress.get_next_step('main'), 'steps':len(svgs_data) })

    cam = get_drawing_camera()
    working_scene = get_working_scene()
    transform_factor = UNITS_SYSTEM[working_scene.scene.unit_settings.system][
            'svg_factor'] * working_scene.resolution_correction
    for i, svg_data in enumerate(svgs_data):
        progress.update(phase_name='collect_subject_svgs', step=i)
        subject = svg_data.subject
        if svg_data.subject not in subjects:
            continue
        layers_labels = [drawing_styles[d_style].name for d_style in \
                sorted_dict(drawing_styles)]
        abs_svg = get_abs_svg(get_working_scene().svg_size, 
                get_css_style('../'), layers_labels)
        for svg_key in svg_data.data:
            ## Get the polylines coords from styled svg 
            ## and multiply them for the svg_factor
            all_points = []
            key_svg_drawing = svg_data.data[svg_key]['svg_content'].drawing
            polylines = key_svg_drawing.get_entities_by_tag('polyline', True)
            for pl in polylines:
                pl.points = transform_points(pl.points, transform_factor, 
                        ROUNDING)
                all_points.append(pl.points[:])
            pl_coords = join_coords(all_points) if svg_key == 'cut' \
                    else [pl.points for pl in polylines]

            svg_data.data[svg_key]['coords'] = pl_coords

            ## Create the paths and the css classes
            abspaths = [AbsPath(coords_string = get_path_coords(coords), 
                coords_values = coords) for coords in pl_coords]
            subj_classes = [svg_key] + subject.get_svg_classes()

            ## Create the group for the style and add paths and classes to it
            absgroup = AbsGroup(entity_id = subject.get_svg_id(svg_key))
            absgroup.add_class(svg_key)
            absgroup.add_class(cam.name)
            for abspath in abspaths:
                abspath.add_class(cam.name) 
                for subj_class in subj_classes:
                    abspath.add_class(subj_class)
                ## Add all the paths to the layer group
                absgroup.add_entity(abspath)

            ## Get the corresponding layer and add the group to it
            svg_layers = abs_svg.get_entities_by_tag('g', True)
            abslayer = [i for i in svg_layers if hasattr(i, 'label') and 
                    i.label == svg_key][0]
            abslayer.add_entity(absgroup)
            abslayer.add_class(svg_key)
            abslayer.add_class(cam.name)

        ## Make the abssvg real and add subject data to it
        subj_svg = abs_svg.to_real(svg_data.path)
        with open(svg_data.path, "a") as svg_file:
            append_subject_data(svg_file, svg_data.subject)
    print(f'\t...completed in {(time.time() - rewrite_svgs_start_time)}\n')

def append_subject_data(svg_file: 'io.TextIOWrapper', 
        subject: 'Drawing_subject') -> None:
    """ Append data about subject in svg_file """
    svg_file.write("<!--" + os.linesep)
    working_scene = get_working_scene()
    svg_file.write(f'{{')
    svg_file.write(f'"subject": "{subject.name}",')
    svg_file.write(os.linesep)
    svg_file.write(f'"original_name": "{subject.original_name}",')
    svg_file.write(os.linesep)
    svg_file.write(f'"resolution": {working_scene.get_resolution()},')
    for cond in subject.conditions:
        svg_file.write(os.linesep)
        svg_file.write(f'"{cond}": "{subject.conditions[cond]}",')
    svg_file.write(os.linesep)
    svg_file.write('"overlaps": [')
    for over_subj in subject.overlapping_subjects:
        lib = str(over_subj.lib_path) if over_subj.lib_path else None
        svg_file.write(f'("{over_subj.name}", "{lib}"),')
    svg_file.write('],')
    svg_file.write(os.linesep)
    ## pixel are collected in ranges with first and last included
    svg_file.write(f'"render_pixels": {subject.pixels_range},')
    svg_file.write(f'}}')
    svg_file.write(os.linesep)
    svg_file.write("-->")

def css_to_inches(css_src: Filepath, css_dest: Filepath) -> None:
    """ Convert millimeter values in css to inches values """
    with open(css_src, encoding='utf-8') as f:
        css_content = f.readlines()
    new_css = []
    for line in css_content:
        stroke_search = re_stroke_width.search(line)
        dasharray_search = re_stroke_dasharray.search(line)
        if not stroke_search and not dasharray_search:
            new_css.append(line)
            continue
        if stroke_search:
            new_stroke_width = float(stroke_search.group(2)) / 25.4
            new_line = re_stroke_width.sub(r'\1 ' + str(new_stroke_width), line)
        elif dasharray_search:
            dasharray_values = dasharray_search.group(2).split()
            new_dasharray = ' '.join([str(float(dash) / 25.4) \
                    for dash in dasharray_values])
            new_line = re_stroke_dasharray.sub(r'\1 ' + new_dasharray, line)
        new_css.append(new_line)
    with open(css_dest, 'w') as f:
        f.write(''.join(new_css))


def get_css(dest_dir: Filepath, inches: bool = False) -> None:
    """ Check if css file is in dest_dir and copy it from ASSETS_DIR if needed 
        Then fix units by calling css_to_inches() """
    css_src = ASSETS_DIR / BASE_CSS
    css_dest = dest_dir / BASE_CSS
    if css_dest.is_file():
        print('CSS is already there. Nothing to add')
        return
    print('CSS is not there: copy it')
    if not inches:
        shutil.copy(css_src, dest_dir)
        return
    css_to_inches(css_src, css_dest)

def get_svg_composition(subjects: list['Drawing_subject'], draw_all: bool) \
        -> "pathlib.PosixPath":
    """ Collect every subject svg in a single composed svg 
        or add new subject to existing composed svg """
    ## TODO check why a lot of defs elements are created
    print('Start composition')
    composition_start_time = time.time()
    layers_labels = [drawing_styles[d_style].name for d_style in \
            sorted_dict(drawing_styles)]

    cam = get_drawing_camera()
    base_path = Filepath(bpy.path.abspath('//'))
    composition = base_path / cam.svg_path
    working_scene = get_working_scene()
    print('Filepath:', composition)
    ## TODO try to set cm as display units of svg 
    if not composition.is_file() or draw_all:
        ## Create a new svg and collect all subject in it as use
        abstract_composition = get_abs_svg(working_scene.svg_size, 
                get_css_style(''))
        absoverall_group = AbsLayer(label = cam.name, entity_id = cam.name)
        abstract_composition.add_entity(absoverall_group)
        for layer_label in layers_labels:
            abslayer = AbsLayer(label = layer_label, entity_id = layer_label)
            abslayer.add_class(layer_label)
            abslayer.add_class(cam.name)
            absoverall_group.add_entity(abslayer)
            for subject in subjects:
                subject_use = subject.get_svg_use(layer_label)
                if not subject_use:
                    continue
                abslayer.add_entity(subject_use)
    else:
        existing_composition = Svg_read(composition).drawing
        existing_subjects = subjects_in_svg(existing_composition, subjects)
        new_subjects = []
        for subj in subjects:
            if subj not in existing_subjects:
                new_subjects.append(subj) 
                continue
            for style in subj.styles:
                if style in existing_subjects[subj]:
                    continue
                new_subjects.append(subj)
        ## TODO fix: adding hidden drawing recreate original too 
        ##      (so it appears twice)
        if new_subjects:
            for style in layers_labels:
                svg_layers = existing_composition.get_entities_by_tag('g', True)
                containers = [i for i in svg_layers if hasattr(i, 'label') and 
                        i.label == style]
                if not containers:
                    continue
                for subject in new_subjects:
                    subject_use = subject.get_svg_use(style)
                    if not subject_use:
                        continue
                    containers[0].add_entity(subject_use)
        abstract_composition = existing_composition
    abstract_composition.to_real(composition)
    inches = working_scene.scene.unit_settings.system == 'IMPERIAL'
    get_css(cam.svg_path.parent.absolute(), inches)
    print(f'\t...completed in {(time.time() - composition_start_time)}\n')
    return composition

def process_args() -> tuple:
    """ Get objects and camera from sys.argv or from selection """
    cam, objs = None, []
    args = [arg for arg in sys.argv[sys.argv.index("--") + 1:]]
    print('args', args)
    args_objs_names = ''.join(args).split(';') if args else []
    for obj_name in args_objs_names:
        if bpy.data.objects[obj_name].type == 'CAMERA':
            cam = bpy.data.objects[obj_name]
            continue
        if is_renderable(bpy.data.objects[obj_name]):
            objs.append(bpy.data.objects[obj_name])
    if not cam:
        for obj in bpy.context.selected_objects:
            if obj.type != 'CAMERA':
                continue
            cam = obj
            break
    if not objs:
        for obj in bpy.context.selected_objects:
            if not is_renderable(obj):
                continue
            objs.append(obj)
    return cam, objs

def get_subjects() -> list['Drawing_subject']:
    """ Find all visible subjects (symbols, forced, cut and back subjects 
        are added anyway) and move them to the working scene """
    subjects = []
    working_scene = get_working_scene()
    scene = working_scene.scene
    drawing_camera = get_drawing_camera() 
    ## Get all subjects inside camera frame
    
    in_frame_subjects = get_in_frame_subjects()
    next(main_progress_count) ## -> get_in_frame_subjects @ subject_getter

    not_occluded_subjects = get_not_occluded_subjects(in_frame_subjects, 
            drawing_camera.data, working_scene)

    progress = get_progress_indicator()
    progress.set_phase(phase_data={
        'name': 'get_subjects', 'start':progress.get_last_update(), 
        'end':progress.get_next_step('main'), 'steps':len(in_frame_subjects)})

    ## Remove not visible subjects and add others to working_scene
    for i, subj in enumerate(in_frame_subjects):
        progress.update(phase_name='get_subjects', step=i)
        if to_be_drawn(subj):
            print(subj.name, 'is TO DRAW')
            working_scene.add_subject(subj)
            subjects.append(subj)
            continue
        if subj.is_cut and is_cut(subj.mesh_eval, subj.matrix, 
                drawing_camera.frame, drawing_camera.direction, scene):
            print(subj.name, 'is CUT')
            working_scene.add_subject(subj)
            subjects.append(subj)
            continue
        if subj not in not_occluded_subjects:
            print(subj.name, 'is NOT VISIBLE')
            subj.remove()
            continue
        print(subj.name, 'is VISIBLE')
        working_scene.add_subject(subj)
        subjects.append(subj)
    return subjects

def reset_scene() -> None:
    """ Remove all data created by prj """
    cam = get_drawing_camera()
    if cam:
        cam.remove()
    reset_svgs_data()
    reset_subjects_list()
    get_working_scene().remove(del_subjs=True, clear_objs=True)
    global main_progress_count
    main_progress_count = itertools.count()

def main(cam: bpy.types.Object = None, objs: list[bpy.types.Object] = None,
        context: bpy.types.Context = None) -> "pathlib.PosixPath":

    if inspect.stack()[1].function == '<module>':
        ## main() is launched by terminal (not by UI through __init__.py)
        print('Start now')
        start_time = time.time()

        context = bpy.context
        cam, objs = process_args()
        render_basepath = get_render_basepath()
        if not cam:
            error_message = "Just one camera has to be selected"
            print("\n", error_message)
            return
        if not check_dir(render_basepath):
            error_message = f'Can\'t find output path folder: {render_basepath}'
            print("\n", error_message)
            return

    global drawing_styles
    drawing_styles = get_drawing_styles()

    ## Prepare progress indicator
    with open(__file__, 'r') as f:
        lines = f.read().split('\n')
    main_progresses = [line for line in lines \
            if re.search(r"next\(main_progress_count\)", line)]
    progress = get_progress_indicator()
    progress.set_phase(phase_data={'name': 'main', 'start':0., 'end':1., 
                'steps':len(main_progresses)})
    progress.update(phase_name='main', value=0.)

    resolutions = get_resolution(camera_data = cam.data)
    working_scene = get_working_scene(resolution = resolutions['resolution'],
            raw_resolution = resolutions['raw'],
            resolution_percentage = set_hi_res_percent(cam.data))
    drawing_camera = get_drawing_camera(cam) 
    progress.update(phase_name='main', step=next(main_progress_count))

    subjects = get_subjects()
    next(main_progress_count) ## -> local
    get_overlaps(subjects)
    next(main_progress_count) ## -> get_overlaps @ subject_processor
    next(main_progress_count) ## -> render_separate_subjects @ subject_processor

    subjects_to_draw = get_selected_subjects(subjects, objs)
    context.window.scene = get_working_scene().scene

    prepare_subjects(subjects_to_draw)
    next(main_progress_count) ## -> local
    draw_subjects(subjects_to_draw)
    next(main_progress_count) ## -> local
    
    collect_subject_svgs(subjects_to_draw)
    next(main_progress_count) ## -> local

    filepath = get_svg_composition(subjects_to_draw, not bool(objs))
    progress.update(phase_name='main', step=next(main_progress_count))
    if drawing_camera.export_cad:
        dxf_filepath = filepath.with_suffix('.dxf')

        dxf = generate_dxf(dxf_filepath)
        if drawing_camera.export_dwg:
            addon_preferences = get_addon_preferences()
            oda_file_converter = addon_preferences.oda_fc_path \
                    if addon_preferences else ''
            oda_cmd = f"{oda_file_converter} {filepath.parent} {filepath.parent}"
            oda_cmd += f" ACAD2013 DWG 1 1 {dxf_filepath.name}"
            subprocess.run(oda_cmd, shell=True)
        if not drawing_camera.export_dxf:
            remove_file(dxf_filepath)
    for f in tmp_imgs + tmp_svgs:
        remove_file(f)
    progress.update(phase_name='main', step=next(main_progress_count))

    progress.end()

    if inspect.stack()[1].function == '<module>':
        reset_scene()
        print(f"\n--- Completed in {time.time() - start_time} seconds ---\n\n")

    return filepath

if __name__ == "__main__":
    main()
